
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import AdaBoostRegressor
from sklearn.linear_model import ElasticNet
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import max_error
from wuml.type_check import *
import numpy as np
import pandas as pd
import wuml

class regression:
	'''
	Automatically run regression on data

	data: can be any data type
	regressor='GP', 'linear', 'NeuralNet', 'kernel ridge', 'AdaBoost', 'Elastic net', 'RandomForest', 'Predef_NeuralNet'
				if 'Predef_NeuralNet' is used, you can additionally set arguments : networkStructure, max_epoch, learning_rate
	split_train_test: automatically splits the data, default is true
	alpha: weight of the Kernel Ridge regularizer
	gamma: weight of the Gaussian kernel
	'''

	def __init__(self, data, y=None, y_column_name=None, split_train_test=True, regressor='GP',  
				alpha=0.5, gamma=1, l1_ratio=0.5, network_info_print=False,
				networkStructure=[(100,'relu'),(100,'relu'),(1,'none')], max_epoch=500, learning_rate=0.001	):
		NP = ensure_numpy
		S = np.squeeze

		X = NP(data)
		y = ensure_label(data, y=y, y_column_name=y_column_name)

		if split_train_test:
			self.X_train, self.X_test, self.y_train, self.y_test = wuml.split_training_test(data, label=y, xdata_type="%.4f", ydata_type="%.4f")
		else:
			self.X_train = X
			self.y_train = y

		if regressor == 'GP':
			model = GaussianProcessRegressor(kernel=None, random_state=0)
		elif regressor == 'linear':
			model = LinearRegression()
		elif regressor == 'kernel ridge':
			model = KernelRidge(alpha=alpha, gamma=gamma, kernel='rbf')
		elif regressor == 'AdaBoost':
			model = AdaBoostRegressor(random_state=0, n_estimators=100)
		elif regressor == 'Elastic net':
			model = ElasticNet(random_state=0, alpha=alpha, l1_ratio=l1_ratio)
		elif regressor == 'NeuralNet':
			model = MLPRegressor(random_state=1, max_iter=1000)
		elif regressor == 'RandomForest':
			model = RandomForestRegressor(max_depth=3, random_state=0)
		elif regressor == 'Predef_NeuralNet':
			Xt = ensure_wData(self.X_train)
			model = wuml.basicNetwork('mse', Xt, Y=S(NP(self.y_train)), networkStructure=networkStructure, 
										max_epoch=max_epoch, learning_rate=learning_rate, network_info_print=network_info_print)
		model.fit(NP(self.X_train), S(NP(self.y_train)))
		try: [self.ŷ_train, self.σ] = model.predict(NP(self.X_train), return_std=True)
		except: 
			self.ŷ_train = model.predict(NP(self.X_train))

		self.mse_train = mean_squared_error(S(NP(self.y_train)), self.ŷ_train)

		if split_train_test:
			try: [self.ŷ_test, self.σ_test] = model.predict(NP(self.X_test), return_std=True)
			except: self.ŷ_test = model.predict(NP(self.X_test))
			self.mse_test = mean_squared_error(S(NP(self.y_test)), self.ŷ_test)

		self.split_train_test = split_train_test
		self.model = model
		self.regressor = regressor

	def score(self, score_type='r2'):

		if score_type == 'r2' or score_type == 'all scores':
			self.train_r2_score = r2_score(self.y_train, self.ŷ_train)
			if self.split_train_test: 
				self.test_r2_score = r2_score(self.y_test, self.ŷ_test)

		if score_type == 'mean absolute error' or score_type == 'all scores':
			self.train_avg_abs_Δ = mean_absolute_error(self.y_train, self.ŷ_train)
			if self.split_train_test: 
				self.test_avg_abs_Δ = mean_absolute_error(self.y_test, self.ŷ_test)

		if score_type == 'max error' or score_type == 'all scores':
			self.train_max_abs_Δ = max_error(self.y_train, self.ŷ_train)
			if self.split_train_test: 
				self.test_max_abs_Δ = max_error(self.y_test, self.ŷ_test)



	def result_summary(self, print_out=True):
		NPR = np.round
		self.score(score_type='all scores')

		if self.split_train_test:
			column_names = ['Train mse', 'Test mse', 
							'Train r2 Score', 'Test r2 Score', 
							'Train avg abs error', 'Test avg abs error', 
							'Train max error', 'Test max error' ]

			data = np.array([[	NPR(self.mse_train,4), 
								NPR(self.mse_test,4), 
								NPR(self.train_r2_score,4), 
								NPR(self.test_r2_score,4), 
								NPR(self.train_avg_abs_Δ,4), 
								NPR(self.test_avg_abs_Δ,4), 
								NPR(self.train_max_abs_Δ,4), 
								NPR(self.test_max_abs_Δ, 4) ]])
		else:
			column_names = ['Train mse', 'Train r2 Score',											
											'Train avg abs error', 'Train max error']

			data = np.array([[NPR(self.mse_train,4), 
							  NPR(self.train_r2_score,4), 
							  NPR(self.train_avg_abs_Δ,4), 
							  NPR(self.train_max_abs_Δ,4)]])

		df = pd.DataFrame(data, columns=column_names,index=[self.regressor])
		if print_out: print(df)
		return df



	def __call__(self, data):
		X = ensure_numpy(data)	

		try:
			[self.ŷ, self.σ] = self.model.predict(X, return_std=True, return_cov=False)
		except:
			self.ŷ = self.model.predict(X)

		return ensure_data_type(self.ŷ, type_name=type(data).__name__)

		#else: raise ValueError('Regressor not recognized, must use regressor="GP"')
		

	def __str__(self):
		return str(self.result_summary(print_out=False))


	def show_true_v_predicted(self, sort_by='Error', ascending=False):
		'''
			sort_by = 'Value' or 'Error'
		'''
		NP = ensure_numpy

		Y = ensure_wData(self.y_train)
		train_E = np.absolute(NP(self.y_train) - NP(self.ŷ_train))
		Y.append_columns(self.ŷ_train)
		Y.append_columns(train_E)
		Y.rename_columns(['Train y','Train ŷ', 'Train Error'])
		if sort_by == 'Value':
			Y.sort_by('Train y', ascending=ascending)
		elif sort_by == 'Error':
			Y.sort_by(['Train Error'], ascending=ascending)


		if self.split_train_test:
			Yt = ensure_wData(self.y_test)
			test_E = np.absolute(NP(self.y_test) - NP(self.ŷ_test))

			Yt.append_columns(self.ŷ_test)
			Yt.append_columns(test_E)
			Yt.rename_columns(['Test y', 'Test ŷ', 'Test Error'])

			if sort_by == 'Value':
				Yt.sort_by('Test y', ascending=ascending)
			elif sort_by == 'Error':
				Yt.sort_by('Test Error', ascending=ascending)

			padding_len = Y.shape[0] - Yt.shape[0]
			an_array = np.empty((padding_len, 3))
			an_array[:] = np.NaN
			padding = pd.DataFrame(an_array)
			Yt.append_rows(padding)

			Y.reset_index()
			Yt.reset_index()
			Y.append_columns(Yt)
		return Y

def run_every_regressor(data, y=None, y_column_name=None, order_by='Test mse', ascending=True, 
						alpha=1, gamma=1, l1_ratio=0.2, max_epoch=500, learning_rate=0.001,
						network_info_print=False):
	'''
	order_by: 'Train mse', 'Test mse', 'Train r2 Score', 'Test r2 Score', 
				'Train avg abs error', 'Test avg abs error', 'Train max error', 'Test max error'
	'''
	regressors=['Elastic net', 'linear', 'kernel ridge', 'AdaBoost', 'GP', 'NeuralNet', 'RandomForest', 'Predef_NeuralNet']
	results = {}

	df = pd.DataFrame()
	for reg in regressors:
		results[reg] = regression(data, y=y, regressor=reg, alpha=alpha, gamma=gamma, 
							l1_ratio=l1_ratio, max_epoch=max_epoch, learning_rate=learning_rate,
							network_info_print=network_info_print)
		df = df.append(results[reg].result_summary(print_out=False))

	results['Train/Test Summary'] = df.sort_values(order_by, ascending=ascending)
	return results




#!/usr/bin/env python

import os 
import sys

if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')
if os.path.exists('/home/chieh/code/wuML'):
	sys.path.insert(0,'/home/chieh/code/wuML')

import types
import torch
import pickle
import numpy as np
import wuml

import wplotlib
import pandas as pd
from pathlib import Path
from sklearn.metrics import accuracy_score
from IPython.display import clear_output, HTML, Math

def jupyter_print(value, display_all_rows=False, display_all_columns=False, font_size=3, latex=False):
	#font_size is from 1 to 6
	font_size = int(6 - font_size)
	if font_size > 6: font_size = 6
	if font_size < 0: font_size = 1

	if wuml.isnotebook():
		if wuml.wtype(value) == 'DataFrame': 
			if display_all_rows: pd.set_option('display.max_rows', None)
			if display_all_columns: pd.set_option('display.max_columns', None) 
			display(value)
		elif wuml.wtype(value) == 'wData': 
			if display_all_rows: pd.set_option('display.max_rows', None)
			if display_all_columns: pd.set_option('display.max_columns', None)
			display(value.df)
		elif wuml.wtype(value) == 'Index': 
			value = str(value.tolist())
			str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
			display(HTML(data=str2html))
		elif wuml.wtype(value) == 'tuple': 
			value = str(value)
			str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
			display(HTML(data=str2html))
		elif wuml.wtype(value) == 'str': 
			if latex:
				display(Math(r'%s'%value))
			else:
				str2html = '<html><body><h%d>%s</h%d></body></html>'%(font_size, value, font_size)
				display(HTML(data=str2html))
		else:
			print(value)

		pd.set_option('display.max_rows', 10)
	else:
		print(value)



def set_terminal_print_options(precision=3):
	np.set_printoptions(precision=precision)
	np.set_printoptions(linewidth=400)
	np.set_printoptions(suppress=True)

	torch.set_printoptions(edgeitems=3)
	torch.set_printoptions(sci_mode=False)
	torch.set_printoptions(precision=precision)
	torch.set_printoptions(linewidth=400)

	pd.set_option("display.max_rows", 500)
	pd.set_option("precision", precision)

def csv_out(X, path, add_row_indices=False, include_column_names=False, float_format='%.4f'):
	X = wuml.ensure_wData(X)
	X.to_csv(path, add_row_indices=add_row_indices, include_column_names=include_column_names, float_format=float_format)

def read_txt_file(path):
	f = open(path, 'r')
	content = f.read()
	f.close()
	return content

def pickle_dump(obj, path):
	pickle.dump( obj, open( path, "wb" ) )
   
def pickle_load(pth):
	return pickle.load( open( pth, "rb" ) )

def copy_file(src, destination):
	copyfile(src, destination)

def write_to(txt, path):
	f = open(path, 'w')
	f.write(txt)
	f.close()

def go_up_1_directory(full_path):	# path should be a string
	return str(Path(full_path).parents[0])

def initialize_empty_folder(path):
	ensure_path_exists(path)
	remove_files(path)

def path_exists(path):
	if os.path.exists(path): return True
	return False

def ensure_path_exists(path):
	if os.path.exists(path): return True
	os.mkdir(path)
	return False

def path_list_exists(path_list):
	for i in path_list:
		if os.path.exists(i) == False: 
			return False

	return True

def create_file(path):
	fin = open(path,'w')
	fin.close()

def delete_file(path):
	if os.path.exists(path):
		os.remove(path)

def remove_files(folder_path):	# make sure to end the path with /
	if not os.path.exists(folder_path): return 

	file_in_tmp = os.listdir(folder_path)
	for i in file_in_tmp:
		if os.path.isfile(folder_path + i):
			os.remove(folder_path + i)





#	Terminal Printing

def clear_current_line():
	sys.stdout.write("\r")
	sys.stdout.write("\033[K")
	sys.stdout.flush()

def write_to_current_line(txt):
	clear_current_line()
	clear_output(wait = True)
	sys.stdout.write(txt)

def clear_previous_line():
	clear_current_line()
	sys.stdout.write("\r")
	sys.stdout.write("\033[F")
	sys.stdout.flush()
	clear_current_line()

def loss_optimization_printout(db, epoch, avgLoss, avgGrad, epoc_loop, slope):
	sys.stdout.write("\r\t\t%d/%d, MaxLoss : %f, AvgGra : %f, progress slope : %f" % (epoch, epoc_loop, avgLoss, avgGrad, slope))
	sys.stdout.flush()

def dictionary_to_str(dic):
	outstr = ''
	for i,j in dic.items():
		if type(j) == str: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == np.float64: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == bool: outstr += ('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == type: outstr += ('\t\t' + i + ' : ' + j.__name__ + '\n')
		elif type(j) == types.FunctionType: outstr += ('\t\t' + i + ' : ' + j.__name__ + '\n')
		elif type(j) == float: outstr += ('\t\t' + i + ' : ' + str(j) + '\n')
		elif type(j) == int: outstr += 	('\t\t' + i + ' : ' + str(j) + '\n')
		else:
			print('%s , %s is not recognized'%(i, str(type(j))))
			import pdb; pdb.set_trace()	

	return outstr

def output_two_columns_side_by_side(col_1, col_2, labels=None, rounding=3):
	col_1 = np.atleast_2d(wuml.ensure_numpy(col_1, rounding=rounding))
	col_2 = np.atleast_2d(wuml.ensure_numpy(col_2, rounding=rounding))

	if col_1.shape[0] == 1: col_1 = col_1.T
	if col_2.shape[0] == 1: col_2 = col_2.T

	output = wuml.pretty_np_array(np.hstack((col_1, col_2)))
	if labels is not None:
		output = wuml.pretty_np_array(labels) + output

	return output

def output_regression_result(y, ŷ, write_path=None):
	y = np.atleast_2d(wuml.ensure_numpy(y, rounding=2))
	ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=2))

	if y.shape[0] == 1: y = y.T
	if ŷ.shape[0] == 1: ŷ = ŷ.T
	
	# Draw Histogram
	Δy = np.absolute(ŷ - y)
	avg_Δ = 'Avg error: %.4f\n\n'%(np.sum(Δy)/Δy.shape[0])

	H = wplotlib.histograms()
	H.histogram(Δy, num_bins=15, title='Histogram of Errors', 
				xlabel='Error Amount', ylabel='Error count',
				facecolor='blue', α=0.5, path=None)


	A = wuml.pretty_np_array(np.array([['y', 'ŷ']]))
	B = wuml.pretty_np_array(np.hstack((y, ŷ)))
	C = avg_Δ + A + B

	if write_path is not None: wuml.write_to(C, write_path)

	return C


class summarize_regression_result:
	def __init__(self, y, ŷ):
		y = np.atleast_2d(wuml.ensure_numpy(y, rounding=2))
		ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=2))
	
		if y.shape[0] == 1: y = y.T
		if ŷ.shape[0] == 1: ŷ = ŷ.T
		
		self.y = y
		self.ŷ = ŷ
		self.side_by_side_Y = np.hstack((self.y, self.ŷ))

		self.Δy = np.absolute(self.ŷ - self.y)

	def avg_error(self):
		Δy = self.Δy
		avg_Δ = np.sum(Δy)/Δy.shape[0]
		return avg_Δ

	def error_histogram(self):
		Δy = self.Δy
		avg_Δ = 'Avg error: %.4f\n\n'%(np.sum(Δy)/Δy.shape[0])
	
		H = wplotlib.histograms()
		H.histogram(Δy, num_bins=15, title='Histogram of Errors', 
					xlabel='Error Amount', ylabel='Error count',
					facecolor='blue', α=0.5, path=None)
		
	def true_vs_predict(self, write_path=None, sort_based_on_label=False, print_result=False):
		#A = wuml.pretty_np_array(np.array([['y', 'ŷ']]))
		A = np.array([['y', 'ŷ']])
		if sort_based_on_label:
			Yjoing = np.hstack((self.y, np.round(self.ŷ,3)))
			sorted_df = wuml.sort_matrix_rows_by_a_column(Yjoing, 0)
			#B = wuml.pretty_np_array(sorted_df.values)
			B = sorted_df.values
		else: 
			B = np.hstack((self.y, self.ŷ))

		return wuml.ensure_wData(np.vstack((A,B)))


		#import pdb; pdb.set_trace()

		#avg_Δ = 'Avg error: %.4f\n\n'%(self.avg_error())
		#C = avg_Δ + A + B
	
		#if print_result: print(C)
		#if write_path is not None: wuml.write_to(C, write_path)

		#return ensure_wData(C, column_names=None)
		#return C
	

class summarize_classification_result:
	def __init__(self, y, ŷ):
		y = np.atleast_2d(wuml.ensure_numpy(y, rounding=2))
		ŷ = np.atleast_2d(wuml.ensure_numpy(ŷ, rounding=2))
	
		if y.shape[0] == 1: y = y.T
		if ŷ.shape[0] == 1: ŷ = ŷ.T
		
		self.y = y
		self.ŷ = ŷ
		self.side_by_side_Y = np.hstack((self.y, self.ŷ))

		self.Δy = np.absolute(self.ŷ - self.y)

	def avg_error(self):
		Acc= accuracy_score(self.y, self.ŷ)
		return Acc

	def error_histogram(self):
		pass

		#Δy = self.Δy
		#avg_Δ = 'Avg error: %.4f\n\n'%(np.sum(Δy)/Δy.shape[0])
	
		#H = wplotlib.histograms()
		#H.histogram(Δy, num_bins=15, title='Histogram of Errors', 
		#			xlabel='Error Amount', ylabel='Error count',
		#			facecolor='blue', α=0.5, path=None)
		
	def true_vs_predict(self, write_path=None, sort_based_on_label=False, print_result=False):
		A = wuml.pretty_np_array(np.array([['y', 'ŷ']]))

		if sort_based_on_label:
			Yjoing = np.hstack((self.y, self.ŷ))
			sorted_df = wuml.sort_matrix_rows_by_a_column(Yjoing, 0)
			B = wuml.pretty_np_array(sorted_df.values)
		else: B = wuml.pretty_np_array(np.hstack((self.y, self.ŷ)))
		avg_Δ = 'Avg error: %.4f\n\n'%(self.avg_error())
		C = avg_Δ + A + B
	
		if print_result: print(C)
		if write_path is not None: wuml.write_to(C, write_path)
		return C
	





import numpy as np; np.random.seed(0)
from wplotlib import lines		#pip install wplotlib
from wplotlib import heatMap
from wplotlib import histograms
from wplotlib import scatter

import sys
import os

if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')



import wuml 
from wuml.data_loading import wData

import io
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns; sns.set_theme()
import itertools
import random


def get_feature_histograms(X, path=None, title='', ylogScale=False):
	X = wuml.ensure_numpy(X)

	if path is not None:
		header = './results/DatStats/'
		wuml.ensure_path_exists('./results')
		wuml.ensure_path_exists(header)

	H = histograms(X, num_bins=10, title=title, fontsize=12, facecolor='blue', α=0.5, path=path, subplot=None, ylogScale=ylogScale)

def identify_missing_data_per_feature(data):
	df = wuml.ensure_DataFrame(data)

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	mdp = missing_data_counter_per_feature = []
	for column in df:
		x = df[column].values
		missing_percentage = np.sum(np.isnan(x))/n
		mdp.append(missing_percentage)

	return mdp

def missing_data_stats(data, save_plots=False): 
	df = wuml.ensure_DataFrame(data)

	if save_plots:
		header = './results/DatStats/'
		wuml.ensure_path_exists('./results')
		wuml.ensure_path_exists(header)
	
		mdp = identify_missing_data_per_feature(df)
		mdp = np.array(mdp)
		textstr = ''
		x = np.arange(1, len(mdp)+1)
	
		lp = lines(x, mdp, 'Missing Percentage', 'Feature ID', 'Percentage Missing', subplot=121,
					imgText=textstr, outpath= header + 'feature_missing_percentage.png', 
					xtick_locations=x, xtick_labels=df.columns.to_numpy(), xticker_rotate=30)

		X2 = np.isnan(df.values).astype(int)
		hMap = heatMap()
		hMap.draw_HeatMap(X2, title='Missing Data Heat Map', subplot=122,
								xlabel='Feature ID', ylabel='Sample ID',
								path= header + 'missing_data_heatMap.png')
	
		buffer = io.StringIO()
		df.info(buf=buffer, verbose=True)
		s = buffer.getvalue()
		wuml.write_to(s, header + 'feature_stats.txt')
	else:
		mdp = identify_missing_data_per_feature(df)
		mdp = np.array(mdp)
		textstr = ''
		x = np.arange(1, len(mdp)+1)
	
		xtick_locations=x
		column_names = df.columns.to_numpy()
		if len(column_names) > 10: 
			column_names = None
			xtick_locations = None

		lp = lines(x, mdp, 'Missing Percentage', 'Feature ID', 'Percentage Missing', imgText=textstr, subplot=211,
					xtick_locations=xtick_locations, xtick_labels=column_names, xticker_rotate=30, figsize=(8,12))	

		#	Show heatmap
		X2 = np.isnan(df.values).astype(int)
		xtick_locations = (x-0.5)
		if column_names is None: xtick_locations = None
		heatMap(X2, title='Missing Data Heat Map', subplot=212, xlabel='Feature ID', xticker_rotate=40, 
					ylabel='Sample ID', xtick_locations=xtick_locations, xtick_labels=column_names)
		lp.show()

		return mdp

def get_redundant_pairs(df):
	'''Get diagonal and lower triangular pairs of correlation matrix'''
	pairs_to_drop = set()
	cols = df.columns
	for i in range(0, df.shape[1]):
		for j in range(0, i+1):
			pairs_to_drop.add((cols[i], cols[j]))
	return pairs_to_drop


def get_top_abs_correlations(df, n=5):
	#au_corr = df.corr().abs().unstack()
	au_corr = df.unstack()
	labels_to_drop = get_redundant_pairs(df)
	au_corr = au_corr.drop(labels=labels_to_drop).sort_values(ascending=False, key=abs)
	return au_corr[0:n]

def feature_wise_correlation(data, n=10, label_name=None, get_top_corr_pairs=False, num_of_top_dependent_pairs_to_plot=0):
	'''
		if label_name is label string, then it only compares the features against the label
		num_of_top_dependent_pairs_to_plot: if > 0, it will plot out the most correlated pairs
	'''

	df = wuml.ensure_DataFrame(data)
	if n > df.shape[1]: n = df.shape[1]

	if label_name is not None and label_name not in df.columns:
		raise ValueError('Error : %s is an unrecognized column name. \nThe list of names are %s'%(label_name, str(df.columns)))

	corrMatrix = df.corr()
	topCorr = get_top_abs_correlations(corrMatrix, n=n).to_frame()
	if not get_top_corr_pairs and label_name not in df.columns: 
		outDF = corrMatrix
	elif label_name is None:
		outDF = topCorr
	else:
		corrVector = corrMatrix[label_name].to_frame()
		topCorr = corrVector.sort_values(label_name, key=abs, ascending=False)
		topCorr = topCorr[1:n]
		outDF = topCorr

	if num_of_top_dependent_pairs_to_plot>0:
		subTopCorr = topCorr.head(num_of_top_dependent_pairs_to_plot)
		idx = subTopCorr.index
		idx = idx.to_frame().values
		
		for i, item in enumerate(idx):
			if len(item) == 1:
				α = item
				β = label_name
				A = df[α].to_numpy()
				B = df[label_name].to_numpy()

			elif len(item) == 2:
				α, β = item
			
				A = df[α].to_numpy()
				B = df[β].to_numpy()
	
			corV = subTopCorr.values[i]
			textstr = r'Order ID: %d, Correlation : %.3f' % (i+1, corV)

			lp = scatter(A, B, α + ' vs ' + β, α, β, imgText=textstr, figsize=(10,5))		# (width, height)


	return wuml.ensure_data_type(outDF, type_name=type(data).__name__)


def feature_wise_HSIC(data, n=10, label_name=None, get_top_dependent_pairs=False, num_of_top_dependent_pairs_to_plot=0):
	'''
		if label_name is label string, then it only compares the features against the label
		num_of_top_dependent_pairs_to_plot: if > 0, it will plot out the most correlated pairs
	'''
	X = wuml.ensure_numpy(data)
	df = wuml.ensure_DataFrame(data)
	d = X.shape[1]

	if n > df.shape[1]: n = df.shape[1]
	if label_name is not None and label_name not in df.columns:
		raise ValueError('Error : %s is an unrecognized column name. \nThe list of names are %s'%(label_name, str(df.columns)))


	lst = list(range(d))
	pair_order_list = itertools.combinations(lst,2)
	depMatrix = np.eye(d)
	for α, β in list(pair_order_list):
		x1 = np.atleast_2d(X[:,α]).T
		x2 = np.atleast_2d(X[:,β]).T

		joinX = wuml.ensure_DataFrame(np.hstack((x1,x2)))
		joinX = joinX.dropna()
		withoutNan = joinX.values
		depMatrix[α,β] = depMatrix[β,α] = wuml.HSIC(withoutNan[:,0], withoutNan[:,1], sigma_type='mpd')

	depM_DF = wuml.ensure_DataFrame(depMatrix, columns=df.columns, index=df.columns)
	topCorr = get_top_abs_correlations(depM_DF, n=n).to_frame()
	if not get_top_dependent_pairs and label_name not in df.columns: 
		outDF = depM_DF
	elif label_name is None:
		outDF = topCorr
	else:
		corrVector = depM_DF[label_name].to_frame()
		topCorr = corrVector.sort_values(label_name, key=abs, ascending=False)
		topCorr = topCorr[1:n]
		outDF = topCorr

	if num_of_top_dependent_pairs_to_plot>0:
		subTopCorr = topCorr.head(num_of_top_dependent_pairs_to_plot)
		idx = subTopCorr.index
		idx = idx.to_frame().values
		
		for i, item in enumerate(idx):
			if len(item) == 1:
				α = item
				β = label_name
				A = df[α].to_numpy()
				B = df[label_name].to_numpy()

			elif len(item) == 2:
				α, β = item
			
				A = df[α].to_numpy()
				B = df[β].to_numpy()
	
			corV = subTopCorr.values[i]
			textstr = r'Order ID: %d, Correlation : %.3f' % (i+1, corV)
			lp = scatter(A, B, α + ' vs ' + β, α, β, imgText=textstr, figsize=(10,5))		# (width, height)


	return wuml.ensure_data_type(outDF, type_name=type(data).__name__)



def HSIC_of_feature_groups_vs_label_list(data, data_compared_to):
	'''
		Compare the entire "data" to each column of "data_compared_to"
	'''
	
	X = wuml.ensure_numpy(data)
	Ys = wuml.ensure_numpy(data_compared_to)
	Ys_df = wuml.ensure_DataFrame(data_compared_to)
	num_of_Ys = Ys.shape[1]

	hsic_list = []
	for i in range(num_of_Ys):
		hsic_list.append(wuml.HSIC(X, Ys[:,i])) #, sigma_type='mpd'

	df = wuml.ensure_DataFrame(np.array(hsic_list))
	df.index = Ys_df.columns
	df.columns = ['feature_group']
	df = df.sort_values('feature_group', axis=0, ascending=False)
	
	return wuml.ensure_wData(df)





import shap		#pip install shap
import xgboost
import sys
import wuml
import numpy as np

class explainer():
	def __init__(self, data, explainer_algorithm='shap', link='identity',
						model=None, 										# Use this if you have your own model, input/output should be nparray
						load_network_path=None,								# You can load your own pytorch network
						loss='mse',  										#use these options if you use a default regression
						networkStructure=[(100,'relu'),(100,'relu'),(1,'none')], 
						max_epoch=1000, learning_rate=0.001, print_network_training_status=False):

		'''
			data : must be wData type with data.X and data.Y defined
			loss: Can be str : 'mse', 'L1', 'CE', 'hindge' or you can define your own loss function
			explainer_algorithm: 'shap', 'XGBRegressor'
			use link='identity' for regression and link='logit' for classification
		'''

		if type(data).__name__ != 'wData': 
			raise ValueError('Error: data input into explainer must be wData type with data.X, data.Y defined.')

		if explainer_algorithm == 'shap':
			X = wuml.ensure_numpy(data)
			if load_network_path is not None:	
				if type(load_network_path) != type(''): 
					print('\nError : load_network_path is a string that you load the network, it must be a basicNetwork type!!!')
					print('\tIf you want to input a function, use model=yourFunction argument instead!!!\n')
					sys.exit()
					
				self.model = wuml.pickle_load(load_network_path)
				if data.label_type == 'continuous': 
					self.model.eval(output_type='ndarray')
					self.Explr = shap.KernelExplainer(self.model, np.zeros((1, X.shape[1])), link=link)		
				else: 
					self.model.eval(output_type='ndarray', out_structural='softmax')
					self.Explr = shap.KernelExplainer(self.model, X, link=link)		
			elif model is not None:
				self.model = model
				self.Explr = shap.KernelExplainer(model, np.zeros((1, X.shape[1])), link=link)
			else:
				self.model = wuml.basicNetwork(loss, data, networkStructure=networkStructure, max_epoch=max_epoch, learning_rate=learning_rate)
				self.model.train(print_status=print_network_training_status)
				if data.label_type == 'continuous': 
					self.model.eval(output_type='ndarray')		# If classification, output softmax
					self.Explr = shap.KernelExplainer(self.model, np.zeros((1, X.shape[1])), link=link)		#, l1_reg=False
				else: 
					self.model.eval(output_type='ndarray', out_structural='softmax')
					self.Explr = shap.KernelExplainer(self.model, X, link=link)		#, l1_reg=False


		elif explainer_algorithm == 'XGBRegressor':
			self.model = xgboost.XGBRegressor().fit(data.df, data.Y)
			self.Explr = shap.Explainer(self.model)


		self.explainer_algorithm = explainer_algorithm


#	def force_plot(self, single_shape_row, X):
#		shap.force_plot(self.Explr.expected_value[0], single_shape_row, X, link="identify")
#
#	def waterfall(self, single_shape_row):
#		shap.plots.waterfall(single_shape_row)

	def __call__(self, data, nsamples=20, output_all_results=False):
		A = self.explainer_algorithm
		if A == 'shap':
			X = wuml.ensure_numpy(data)
			shapOut = self.Explr.shap_values(X, nsamples=nsamples)[0]
			return wuml.ensure_DataFrame(shapOut)

		elif A == 'XGBRegressor':
			X = wuml.ensure_DataFrame(data)
			results = self.Explr(X)
			if output_all_results:
				return results
			else:
				return wuml.ensure_DataFrame(results.values)






from scipy.integrate import quad
from numpy import linalg as LA
import numpy as np
import wuml


def integrate(foo, x0, x1):
	[result, error] = quad(foo, x0, x1)
	return [result, error]


class eig:
	def __init__(self, M, percentage_to_keep=None, num_eigs_to_keep=None, largest_to_smallest=True,
					return_eigen_values_as_diagonal=True, return_contribution_of_each_vector=False):
		'''
			X = VΛVᵀ
				v0 = wuml.get_np_column(V, 0)
				λ0 = Λ[0]
				λ0*v0 == X.dot(v0)
		'''

		X = wuml.ensure_numpy(M)
	
		if X.shape[0] == X.shape[1]: 		# ensure matrix is square
	
			if np.all(np.abs(X-X.T) < 1e-8):	# matrix is symmetric
				Λ,V = np.linalg.eigh(X)

				if largest_to_smallest:
					Λ = np.flip(Λ)
					V = np.flip(V, axis=1)		# X = VΛVᵀ

				cs = np.cumsum(Λ)/np.sum(Λ)
				if percentage_to_keep is not None:	
					residual_ð = np.sum(cs < percentage_to_keep) + 1
				
				elif num_eigs_to_keep is not None:
					residual_ð = num_eigs_to_keep
				else:
					residual_ð = X.shape[0]

				self.eigen_values = Λ[0:residual_ð]
				self.eigen_vectors = V[:, 0:residual_ð]
				self.cumsum_eigen_values = cs
				self.normalized_eigen_values = Λ/np.sum(Λ)
				self.all_eigen_values = Λ
				self.all_eigen_vectors = V			# each columns is an eigenvector

				if return_eigen_values_as_diagonal:
					self.all_eigen_values = np.diag(Λ)
					self.eigen_values = np.diag(self.eigen_values)

				if return_contribution_of_each_vector:
					self.data_projected_onto_eigvectors = V.T.dot(X)

			else:
				raise ValueError('I have not yet written code to handle non-symmetric matrices for eig')
		else:
			raise ValueError('I have not yet written code to handle non-squared matrices for eig')

	def __str__(self):
		return str(self.eigen_vectors)
